/* This file is Copyright 2005 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <stdio.h>

#include "dataio/TCPSocketDataIO.h"
#include "dataio/RS232DataIO.h"

#include "system/SetupSystem.h"
#include "util/NetworkUtilityFunctions.h"
#include "util/StringTokenizer.h"

USING_NAMESPACE(muscle);


static void DoSession(DataIO & io)
{
   SetSocketBlockingEnabled(STDIN_FILENO, false);
   int ss = io.GetSelectSocket();

   while(true)
   {
      fd_set readSet, writeSet;

      FD_ZERO(&readSet);
      FD_SET(ss, &readSet);
      FD_SET(STDIN_FILENO, &readSet);

      char buf[2048];
      if (select(muscleMax(STDIN_FILENO, ss)+1, &readSet, NULL, NULL, NULL) >= 0)
      {
         if (FD_ISSET(ss, &readSet))
         {
            int32 ret = io.Read(buf, sizeof(buf));
            if (ret > 0)
            {
               printf("Received: [");
               for (int32 i=0; i<ret; i++) printf("%s%02x", (i>0)?" ":"", ((uint32)buf[i])&0xFF);
               printf("]\n");
            }
            else if (ret < 0) break;
         }
         if (FD_ISSET(STDIN_FILENO, &readSet))
         {
            fgets(buf, sizeof(buf), stdin);

            uint32 count = 0;
            StringTokenizer tok(false, buf, ", \t\r\n");
            const char * next;
            while((next = tok()) != NULL) 
            {
               if (strlen(next) > 0) buf[count++] = (uint8) strtol(next, NULL, 16);
            }

            if (count>0)
            {
               if (io.Write(buf, count) == count)
               {
                  printf("Sent: [");
                  for (uint32 i=0; i<count; i++) printf("%s%02x", (i>0)?" ":"", ((uint32)buf[i])&0xFF); printf("]\n");
               }
               else break;
            }
         }
      }
      else break;
   }
}

// This program accepts incoming TCP connections on a given port and prints out
// any bytes it receives in Hex format.  Good for debugging clients.
int main(int argc, char ** argv) 
{
   CompleteSetupSystem css;

   if (argc < 2)
   {
      printf("Usage:  listener <port>\n");
      printf("   or:  listener <address> <port>\n");
      exit(5);
   }

   const char * serial = (argc>1)?strstr(argv[1], "serial="):NULL;
   if (serial)
   {
      uint32 baudRate = (argc>2) ? atoi(argv[2]) : 0;
      if (baudRate <= 0) baudRate = 38400;

      String serName;

      serial += 7;
      Queue<String> devs;
      if (RS232DataIO::GetAvailableSerialPortNames(devs) == B_NO_ERROR)
      {
         for (int32 i=devs.GetNumItems()-1; i>=0; i--)
         {
            if (devs[i] == serial) 
            {
               serName = devs[i];
               break;
            }
         }
         if (serName.Length() > 0)
         {
            RS232DataIO io(serial, baudRate, false);
            if (io.IsPortAvailable())
            {
               LogTime(MUSCLE_LOG_INFO, "Communicating with serial port %s (baud rate %lu)\n", serial, baudRate);
               DoSession(io);
               LogTime(MUSCLE_LOG_INFO, "Serial session aborted, exiting.\n");
            }
            else LogTime(MUSCLE_LOG_CRITICALERROR, "Unable to open serial device %s (baud rate %lu).\n", serial, baudRate);
         }
         else 
         {
            LogTime(MUSCLE_LOG_CRITICALERROR, "Serial device %s not found.\n", serial);
            LogTime(MUSCLE_LOG_CRITICALERROR, "Available serial devices are:\n");
            for (uint32 i=0; i<devs.GetNumItems(); i++) LogTime(MUSCLE_LOG_CRITICALERROR, "   %s\n", devs[i]());
         }
      }
      else LogTime(MUSCLE_LOG_CRITICALERROR, "Couldn't get list of serial device names!\n");
   }
   else if (argc >= 3)
   {
      int ss = Connect(argv[1], atoi(argv[2]), "hexterm", false);
      if (ss >= 0) 
      {
         LogTime(MUSCLE_LOG_INFO, "Connected to [%s:%s]\n", argv[1], argv[2]);
         TCPSocketDataIO io(ss, false);
         DoSession(io);
         LogTime(MUSCLE_LOG_INFO, "Session socket disconnected, exiting.\n");
      }
      else LogTime(MUSCLE_LOG_CRITICALERROR, "Unable to connect to [%s:%s]\n", argv[1], argv[2]);
   }
   else
   {
      int p = (argc > 1) ? atoi(argv[1]) : 2980;
      int as = CreateAcceptingSocket(p);
      if (as >= 0)
      {
         LogTime(MUSCLE_LOG_INFO, "Listening for incoming TCP connections on port %i\n", p);
         while(true) 
         {
            int ss = Accept(as);
            if (ss >= 0)
            {
               char buf[64]; Inet_NtoA(GetPeerIPAddress(ss), buf);
               LogTime(MUSCLE_LOG_INFO, "Accepted TCP connection from %s, awaiting data...\n", buf);
               TCPSocketDataIO io(ss, false);
               DoSession(io);
               LogTime(MUSCLE_LOG_ERROR, "Session socket disconnected, awaiting next connection.\n");
            }
         }
         CloseSocket(as);
      }
      else LogTime(MUSCLE_LOG_CRITICALERROR, "Couldn't bind to port %i\n", p);
   }
   return 0;
}
